@@ -28,6 +28,7 @@
 #include <unistd.h>
 #include <string.h>
 #include <errno.h>
+#include <time.h>
 
 #include <glib.h>
 #include <glib/gi18n.h>
@@ -40,10 +41,19 @@
 #include "gnome-settings-profile.h"
 #include "gsd-xsettings-manager.h"
 #include "xsettings-manager.h"
+#ifdef HAVE_FONTCONFIG
+#include "fontconfig-monitor.h"
+#endif /* HAVE_FONTCONFIG */
 
 #define GNOME_XSETTINGS_MANAGER_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GNOME_TYPE_XSETTINGS_MANAGER, GnomeXSettingsManagerPrivate))
 
-#ifdef HAVE_XFT2
+#define MOUSE_SETTINGS_DIR     "/desktop/gnome/peripherals/mouse"
+#define GTK_SETTINGS_DIR       "/desktop/gtk"
+#define INTERFACE_SETTINGS_DIR "/desktop/gnome/interface"
+#define SOUND_SETTINGS_DIR     "/desktop/gnome/sound"
+#define GTK_MODULES_DIR        "/apps/gnome_settings_daemon/gtk-modules"
+
+#ifdef HAVE_FONTCONFIG
 #define FONT_RENDER_DIR "/desktop/gnome/font_rendering"
 #define FONT_ANTIALIASING_KEY FONT_RENDER_DIR "/antialiasing"
 #define FONT_HINTING_KEY      FONT_RENDER_DIR "/hinting"
@@ -64,7 +74,7 @@
 #define DPI_LOW_REASONABLE_VALUE 50
 #define DPI_HIGH_REASONABLE_VALUE 500
 
-#endif /* HAVE_XFT2 */
+#endif /* HAVE_FONTCONFIG */
 
 typedef struct _TranslationEntry TranslationEntry;
 typedef void (* TranslationFunc) (GnomeXSettingsManager *manager,
@@ -82,6 +92,16 @@ struct _TranslationEntry {
 struct GnomeXSettingsManagerPrivate
 {
         XSettingsManager **managers;
+        guint              notify[6];
+#ifdef HAVE_FONTCONFIG
+        fontconfig_monitor_handle_t *fontconfig_handle;
+#endif /* HAVE_FONTCONFIG */
+};
+
+#define GSD_XSETTINGS_ERROR gsd_xsettings_error_quark ()
+
+enum {
+        GSD_XSETTINGS_ERROR_INIT
 };
 
 static void     gnome_xsettings_manager_class_init  (GnomeXSettingsManagerClass *klass);
@@ -92,6 +112,12 @@ G_DEFINE_TYPE (GnomeXSettingsManager, gnome_xsettings_manager, G_TYPE_OBJECT)
 
 static gpointer manager_object = NULL;
 
+static GQuark
+gsd_xsettings_error_quark (void)
+{
+        return g_quark_from_static_string ("gsd-xsettings-error-quark");
+}
+
 static void
 translate_bool_int (GnomeXSettingsManager *manager,
                     TranslationEntry      *trans,
@@ -182,14 +208,18 @@ static TranslationEntry translations [] = {
         { "/desktop/gnome/interface/icon_theme",             "Net/IconThemeName",       GCONF_VALUE_STRING,   translate_string_string },
         { "/desktop/gnome/interface/file_chooser_backend",   "Gtk/FileChooserBackend",  GCONF_VALUE_STRING,   translate_string_string },
         { "/desktop/gnome/interface/menus_have_icons",       "Gtk/MenuImages",          GCONF_VALUE_BOOL,     translate_bool_int },
+        { "/desktop/gnome/interface/buttons_have_icons",     "Gtk/ButtonImages",          GCONF_VALUE_BOOL,     translate_bool_int },
         { "/desktop/gnome/interface/menubar_accel",          "Gtk/MenuBarAccel",        GCONF_VALUE_STRING,   translate_string_string },
         { "/desktop/gnome/peripherals/mouse/cursor_theme",   "Gtk/CursorThemeName",     GCONF_VALUE_STRING,   translate_string_string },
         { "/desktop/gnome/peripherals/mouse/cursor_size",    "Gtk/CursorThemeSize",     GCONF_VALUE_INT,      translate_int_int },
         { "/desktop/gnome/interface/show_input_method_menu", "Gtk/ShowInputMethodMenu", GCONF_VALUE_BOOL,     translate_bool_int },
         { "/desktop/gnome/interface/show_unicode_menu",      "Gtk/ShowUnicodeMenu",     GCONF_VALUE_BOOL,     translate_bool_int },
+        { "/desktop/gnome/sound/theme_name",                 "Net/SoundThemeName",      GCONF_VALUE_STRING,   translate_string_string },
+        { "/desktop/gnome/sound/event_sounds",               "Net/EnableEventSounds" ,  GCONF_VALUE_BOOL,     translate_bool_int },
+        { "/desktop/gnome/sound/input_feedback_sounds",      "Net/EnableInputFeedbackSounds", GCONF_VALUE_BOOL, translate_bool_int }
 };
 
-#ifdef HAVE_XFT2
+#ifdef HAVE_FONTCONFIG
 static double
 dpi_from_pixels_and_mm (int pixels,
                         int mm)
@@ -395,6 +425,7 @@ child_watch_cb (GPid     pid,
 {
         char *command = user_data;
 
+        gnome_settings_profile_end ("%s", command);
         if (!WIFEXITED (status) || WEXITSTATUS (status)) {
                 g_warning ("Command %s failed", command);
         }
@@ -417,6 +448,7 @@ spawn_with_input (const char *command,
                 return;
         }
 
+        gnome_settings_profile_start ("%s", command);
         error = NULL;
         res = g_spawn_async_with_pipes (NULL,
                                         argv,
@@ -518,7 +550,55 @@ xft_callback (GConfClient           *client,
         }
 }
 
-#endif /* HAVE_XFT2 */
+static void
+fontconfig_callback (fontconfig_monitor_handle_t *handle,
+                     GnomeXSettingsManager       *manager)
+{
+        int i;
+        int timestamp = time (NULL);
+
+        gnome_settings_profile_start (NULL);
+
+        for (i = 0; manager->priv->managers [i]; i++) {
+                xsettings_manager_set_int (manager->priv->managers [i], "Fontconfig/Timestamp", timestamp);
+                xsettings_manager_notify (manager->priv->managers [i]);
+        }
+        gnome_settings_profile_end (NULL);
+}
+
+static gboolean
+start_fontconfig_monitor_idle_cb (GnomeXSettingsManager *manager)
+{
+        gnome_settings_profile_start (NULL);
+
+        manager->priv->fontconfig_handle = fontconfig_monitor_start ((GFunc) fontconfig_callback, manager);
+
+        gnome_settings_profile_end (NULL);
+
+        return FALSE;
+}
+
+static void
+start_fontconfig_monitor (GnomeXSettingsManager  *manager)
+{
+        gnome_settings_profile_start (NULL);
+
+        fontconfig_cache_init ();
+
+        g_idle_add ((GSourceFunc) start_fontconfig_monitor_idle_cb, manager);
+
+        gnome_settings_profile_end (NULL);
+}
+
+static void
+stop_fontconfig_monitor (GnomeXSettingsManager  *manager)
+{
+        if (manager->priv->fontconfig_handle) {
+                fontconfig_monitor_stop (manager->priv->fontconfig_handle);
+                manager->priv->fontconfig_handle = NULL;
+        }
+}
+#endif /* HAVE_FONTCONFIG */
 
 static const char *
 type_to_string (GConfValueType type)
@@ -569,6 +649,196 @@ process_value (GnomeXSettingsManager *manager,
         }
 }
 
+static TranslationEntry *
+find_translation_entry (const char *gconf_key)
+{
+        int i;
+
+        for (i = 0; i < G_N_ELEMENTS (translations); ++i) {
+                if (strcmp (translations[i].gconf_key, gconf_key) == 0) {
+                        return &translations[i];
+                }
+        }
+
+        return NULL;
+}
+
+static void
+xsettings_callback (GConfClient           *client,
+                    guint                  cnxn_id,
+                    GConfEntry            *entry,
+                    GnomeXSettingsManager *manager)
+{
+        TranslationEntry *trans;
+        int               i;
+
+        trans = find_translation_entry (entry->key);
+        if (trans == NULL) {
+                return;
+        }
+
+        process_value (manager, trans, entry->value);
+
+        for (i = 0; manager->priv->managers [i]; i++) {
+                xsettings_manager_set_string (manager->priv->managers [i],
+                                              "Net/FallbackIconTheme",
+                                              "gnome");
+        }
+
+        for (i = 0; manager->priv->managers [i]; i++) {
+                xsettings_manager_notify (manager->priv->managers [i]);
+        }
+}
+
+static gchar *
+get_gtk_modules (GConfClient *client)
+{
+        GSList *entries, *l;
+        GString *mods = g_string_new (NULL);
+
+        entries = gconf_client_all_entries (client, GTK_MODULES_DIR, NULL);
+
+        for (l = entries; l != NULL; l = g_slist_next (l)) {
+                GConfEntry *e = l->data;
+                GConfValue *v = gconf_entry_get_value (e);
+
+                if (v != NULL) {
+                        gboolean enabled = FALSE;
+                        const gchar *key;
+
+                        switch (v->type) {
+                        case GCONF_VALUE_BOOL:
+                                /* simple enabled/disabled */
+                                enabled = gconf_value_get_bool (v);
+                                break;
+
+                        /* due to limitations in GConf (or the client libraries,
+                         * anyway), it is currently impossible to monitor
+                         * arbitrary keys for changes, so these won't update at
+                         * runtime */
+                        case GCONF_VALUE_STRING:
+                                /* linked to another GConf key of type bool */
+                                key = gconf_value_get_string (v);
+                                if (key != NULL) {
+                                        enabled = gconf_client_get_bool (client, key, NULL);
+                                }
+                                break;
+
+                        default:
+                                g_warning ("GConf entry %s has invalid type %s",
+                                           gconf_entry_get_key (e), type_to_string (v->type));
+                        }
+
+                        if (enabled) {
+                                const gchar *name;
+                                name = strrchr (gconf_entry_get_key (e), '/') + 1;
+
+                                if (mods->len > 0) {
+                                        g_string_append_c (mods, ':');
+                                }
+                                g_string_append (mods, name);
+                        }
+                }
+
+                gconf_entry_free (e);
+        }
+
+        g_slist_free (entries);
+
+        return g_string_free (mods, mods->len == 0);
+}
+
+static void
+gtk_modules_callback (GConfClient           *client,
+                      guint                  cnxn_id,
+                      GConfEntry            *entry,
+                      GnomeXSettingsManager *manager)
+{
+        gchar *modules = get_gtk_modules (client);
+        int i;
+
+        if (modules == NULL) {
+                for (i = 0; manager->priv->managers [i]; ++i) {
+                        xsettings_manager_delete_setting (manager->priv->managers [i], "Gtk/Modules");
+                }
+        } else {
+                g_debug ("Setting GTK modules '%s'", modules);
+                for (i = 0; manager->priv->managers [i]; ++i) {
+                        xsettings_manager_set_string (manager->priv->managers [i],
+                                                      "Gtk/Modules",
+                                                      modules);
+                }
+                g_free (modules);
+        }
+
+        for (i = 0; manager->priv->managers [i]; ++i) {
+                xsettings_manager_notify (manager->priv->managers [i]);
+        }
+}
+
+static guint
+register_config_callback (GnomeXSettingsManager  *manager,
+                          GConfClient            *client,
+                          const char             *path,
+                          GConfClientNotifyFunc   func)
+{
+        return gconf_client_notify_add (client, path, func, manager, NULL, NULL);
+}
+
+static void
+terminate_cb (void *data)
+{
+        gboolean *terminated = data;
+
+        if (*terminated) {
+                return;
+        }
+
+        *terminated = TRUE;
+
+        gtk_main_quit ();
+}
+
+static gboolean
+setup_xsettings_managers (GnomeXSettingsManager *manager)
+{
+        GdkDisplay *display;
+        int         i;
+        int         n_screens;
+        gboolean    res;
+        gboolean    terminated;
+
+        display = gdk_display_get_default ();
+        n_screens = gdk_display_get_n_screens (display);
+
+        res = xsettings_manager_check_running (gdk_x11_display_get_xdisplay (display),
+                                               gdk_screen_get_number (gdk_screen_get_default ()));
+        if (res) {
+                g_warning ("You can only run one xsettings manager at a time; exiting");
+                return FALSE;
+        }
+
+        manager->priv->managers = g_new0 (XSettingsManager *, n_screens + 1);
+
+        terminated = FALSE;
+        for (i = 0; i < n_screens; i++) {
+                GdkScreen *screen;
+
+                screen = gdk_display_get_screen (display, i);
+
+                manager->priv->managers [i] = xsettings_manager_new (gdk_x11_display_get_xdisplay (display),
+                                                                     gdk_screen_get_number (screen),
+                                                                     terminate_cb,
+                                                                     &terminated);
+                if (! manager->priv->managers [i]) {
+                        g_warning ("Could not create xsettings manager for screen %d!", i);
+                        return FALSE;
+                }
+        }
+
+        return TRUE;
+}
+
 gboolean
 gnome_xsettings_manager_start (GnomeXSettingsManager *manager,
                                GError               **error)
@@ -579,8 +849,22 @@ gnome_xsettings_manager_start (GnomeXSettingsManager *manager,
         g_debug ("Starting xsettings manager");
         gnome_settings_profile_start (NULL);
 
+        if (!setup_xsettings_managers (manager)) {
+                g_set_error (error, GSD_XSETTINGS_ERROR,
+                             GSD_XSETTINGS_ERROR_INIT,
+                             "Could not initialize xsettings manager.");
+                return FALSE;
+        }
+
         client = gconf_client_get_default ();
 
+        gconf_client_add_dir (client, MOUSE_SETTINGS_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+        gconf_client_add_dir (client, GTK_SETTINGS_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+        gconf_client_add_dir (client, INTERFACE_SETTINGS_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+        gconf_client_add_dir (client, SOUND_SETTINGS_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+        gconf_client_add_dir (client, GTK_MODULES_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+        gconf_client_add_dir (client, FONT_RENDER_DIR, GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+
         for (i = 0; i < G_N_ELEMENTS (translations); i++) {
                 GConfValue *val;
                 GError     *err;
@@ -591,7 +875,7 @@ gnome_xsettings_manager_start (GnomeXSettingsManager *manager,
                                         &err);
 
                 if (err != NULL) {
-                        g_warning ("Error getting value for %s: %s\n",
+                        g_warning ("Error getting value for %s: %s",
                                    translations[i].gconf_key,
                                    err->message);
                         g_error_free (err);
@@ -603,9 +887,38 @@ gnome_xsettings_manager_start (GnomeXSettingsManager *manager,
                 }
         }
 
-#ifdef HAVE_XFT2
+        manager->priv->notify[0] =
+                register_config_callback (manager, client,
+                                          MOUSE_SETTINGS_DIR,
+                                          (GConfClientNotifyFunc) xsettings_callback);
+        manager->priv->notify[1] =
+                register_config_callback (manager, client,
+                                          GTK_SETTINGS_DIR,
+                                          (GConfClientNotifyFunc) xsettings_callback);
+        manager->priv->notify[2] =
+                register_config_callback (manager, client,
+                                          INTERFACE_SETTINGS_DIR,
+                                          (GConfClientNotifyFunc) xsettings_callback);
+        manager->priv->notify[3] =
+                register_config_callback (manager, client,
+                                          SOUND_SETTINGS_DIR,
+                                          (GConfClientNotifyFunc) xsettings_callback);
+
+        manager->priv->notify[4] =
+                register_config_callback (manager, client,
+                                          GTK_MODULES_DIR,
+                                          (GConfClientNotifyFunc) gtk_modules_callback);
+        gtk_modules_callback (client, 0, NULL, manager);
+
+#ifdef HAVE_FONTCONFIG
+        manager->priv->notify[5] =
+                register_config_callback (manager, client,
+                                          FONT_RENDER_DIR,
+                                          (GConfClientNotifyFunc) xft_callback);
         update_xft_settings (manager, client);
-#endif /* HAVE_XFT */
+
+        start_fontconfig_monitor (manager);
+#endif /* HAVE_FONTCONFIG */
 
         g_object_unref (client);
 
@@ -618,6 +931,7 @@ gnome_xsettings_manager_start (GnomeXSettingsManager *manager,
                 xsettings_manager_notify (manager->priv->managers [i]);
         }
 
+
         gnome_settings_profile_end (NULL);
 
         return TRUE;
@@ -626,7 +940,41 @@ gnome_xsettings_manager_start (GnomeXSettingsManager *manager,
 void
 gnome_xsettings_manager_stop (GnomeXSettingsManager *manager)
 {
+        GnomeXSettingsManagerPrivate *p = manager->priv;
+        GConfClient *client;
+        int i;
+
         g_debug ("Stopping xsettings manager");
+
+        if (p->managers != NULL) {
+                for (i = 0; p->managers [i]; ++i)
+                        xsettings_manager_destroy (p->managers [i]);
+
+                g_free (p->managers);
+                p->managers = NULL;
+        }
+
+        client = gconf_client_get_default ();
+
+        gconf_client_remove_dir (client, MOUSE_SETTINGS_DIR, NULL);
+        gconf_client_remove_dir (client, GTK_SETTINGS_DIR, NULL);
+        gconf_client_remove_dir (client, INTERFACE_SETTINGS_DIR, NULL);
+        gconf_client_remove_dir (client, SOUND_SETTINGS_DIR, NULL);
+        gconf_client_remove_dir (client, GTK_MODULES_DIR, NULL);
+#ifdef HAVE_FONTCONFIG
+        gconf_client_remove_dir (client, FONT_RENDER_DIR, NULL);
+
+        stop_fontconfig_monitor (manager);
+#endif /* HAVE_FONTCONFIG */
+
+        for (i = 0; i < G_N_ELEMENTS (p->notify); ++i) {
+                if (p->notify[i] != 0) {
+                        gconf_client_notify_remove (client, p->notify[i]);
+                        p->notify[i] = 0;
+                }
+        }
+
+        g_object_unref (client);
 }
 
 static void
@@ -693,7 +1041,7 @@ gnome_xsettings_manager_dispose (GObject *object)
 static void
 gnome_xsettings_manager_class_init (GnomeXSettingsManagerClass *klass)
 {
-        GObjectClass   *object_class = G_OBJECT_CLASS (klass);
+        GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
         object_class->get_property = gnome_xsettings_manager_get_property;
         object_class->set_property = gnome_xsettings_manager_set_property;
@@ -704,125 +1052,10 @@ gnome_xsettings_manager_class_init (GnomeXSettingsManagerClass *klass)
         g_type_class_add_private (klass, sizeof (GnomeXSettingsManagerPrivate));
 }
 
-static TranslationEntry *
-find_translation_entry (const char *gconf_key)
-{
-        int i;
-
-        for (i =0; i < G_N_ELEMENTS (translations); i++) {
-                if (strcmp (translations[i].gconf_key, gconf_key) == 0) {
-                        return &translations[i];
-                }
-        }
-
-        return NULL;
-}
-
-static void
-xsettings_callback (GConfClient           *client,
-                    guint                  cnxn_id,
-                    GConfEntry            *entry,
-                    GnomeXSettingsManager *manager)
-{
-        TranslationEntry *trans;
-        int               i;
-
-        trans = find_translation_entry (entry->key);
-        if (trans == NULL) {
-                return;
-        }
-
-        process_value (manager, trans, entry->value);
-
-        for (i = 0; manager->priv->managers [i]; i++) {
-                xsettings_manager_set_string (manager->priv->managers [i],
-                                              "Net/FallbackIconTheme",
-                                              "gnome");
-        }
-
-        for (i = 0; manager->priv->managers [i]; i++) {
-                xsettings_manager_notify (manager->priv->managers [i]);
-        }
-}
-
-static void
-register_config_callback (GnomeXSettingsManager  *manager,
-                          const char             *path,
-                          GConfClientNotifyFunc   func)
-{
-        GConfClient *client;
-
-        client = gconf_client_get_default ();
-
-        gconf_client_add_dir (client, path, GCONF_CLIENT_PRELOAD_NONE, NULL);
-        gconf_client_notify_add (client, path, func, manager, NULL, NULL);
-
-        g_object_unref (client);
-}
-
-static void
-terminate_cb (void *data)
-{
-        gboolean *terminated = data;
-
-        if (*terminated) {
-                return;
-        }
-
-        *terminated = TRUE;
-
-        gtk_main_quit ();
-}
-
 static void
 gnome_xsettings_manager_init (GnomeXSettingsManager *manager)
 {
-        GdkDisplay *display;
-        int         i;
-        int         n_screens;
-        gboolean    res;
-        gboolean    terminated;
-
         manager->priv = GNOME_XSETTINGS_MANAGER_GET_PRIVATE (manager);
-
-        display = gdk_display_get_default ();
-        n_screens = gdk_display_get_n_screens (display);
-
-        res = xsettings_manager_check_running (gdk_x11_display_get_xdisplay (display),
-                                               gdk_screen_get_number (gdk_screen_get_default ()));
-        if (res) {
-                g_error ("You can only run one xsettings manager at a time; exiting\n");
-                exit (1);
-        }
-
-        manager->priv->managers = g_new (XSettingsManager *, n_screens + 1);
-
-        terminated = FALSE;
-        for (i = 0; i < n_screens; i++) {
-                GdkScreen *screen;
-
-                screen = gdk_display_get_screen (display, i);
-
-                manager->priv->managers [i] = xsettings_manager_new (gdk_x11_display_get_xdisplay (display),
-                                                                     gdk_screen_get_number (screen),
-                                                                     terminate_cb,
-                                                                     &terminated);
-                if (! manager->priv->managers [i]) {
-                        g_error ("Could not create xsettings manager for screen %d!\n", i);
-                        exit (1);
-                }
-        }
-
-        manager->priv->managers [i] = NULL;
-
-        register_config_callback (manager, "/desktop/gnome/peripherals/mouse", (GConfClientNotifyFunc)xsettings_callback);
-        register_config_callback (manager, "/desktop/gtk", (GConfClientNotifyFunc)xsettings_callback);
-        register_config_callback (manager, "/desktop/gnome/interface", (GConfClientNotifyFunc)xsettings_callback);
-
-#ifdef HAVE_XFT2
-        register_config_callback (manager, FONT_RENDER_DIR, (GConfClientNotifyFunc)xft_callback);
-#endif /* HAVE_XFT2 */
-
 }
 
 static void
